/* IMPORTANT
 * This snapshot file is auto-generated, but designed for humans.
 * It should be checked into source control and tracked carefully.
 * Re-generate by setting TAP_SNAPSHOT=1 and running tests.
 * Make sure to inspect the output below.  Do not ignore changes!
 */
'use strict'
exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d: > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "?",
    "d:",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d: > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^(?!\\.)[^/]$/i,
    /^d:$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d:/ > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "?",
    "d:",
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d:/ > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^(?!\\.)[^/]$/i,
    /^d:$/i,
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d:/x/y/z > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "?",
    "d:",
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //?/d:/x/y/z > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^(?!\\.)[^/]$/i,
    /^d:$/i,
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "host",
    "share",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^host$/i,
    /^share$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share/ > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "host",
    "share",
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share/ > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^host$/i,
    /^share$/i,
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share/x/y/z > default to true 1`] = `
Array [
  Array [
    "",
    "",
    "host",
    "share",
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root //host/share/x/y/z > set explicitly false 1`] = `
Array [
  Array [
    "",
    "",
    /^host$/i,
    /^share$/i,
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d: > default to true 1`] = `
Array [
  Array [
    "d:",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d: > set explicitly false 1`] = `
Array [
  Array [
    /^d:$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d:/ > default to true 1`] = `
Array [
  Array [
    "d:",
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d:/ > set explicitly false 1`] = `
Array [
  Array [
    /^d:$/i,
    "",
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d:/x/y/z > default to true 1`] = `
Array [
  Array [
    "d:",
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`

exports[`test/windows-no-magic-root.ts TAP no magic the root d:/x/y/z > set explicitly false 1`] = `
Array [
  Array [
    /^d:$/i,
    /^x$/i,
    /^y$/i,
    /^z$/i,
  ],
]
`
